﻿//----------------------------------------------------------------- 2013-03-29 ------------
// 電力計 サンプルスケッチ
//-------------------------------------------------------- written by iizuka --------------
#include <LiquidCrystal.h>

//---------------- 環境により変更する項目 -------------------------------------------
#define FREQ    50   // 50Hz
//  #define FREQ 60   // 60Hz
#define CT_TURN                   3000  // ＣＴセンサーのコイル巻き数 CTL-10-CLS は 3000巻き
#define CH1_R_LOW_RESOLUTION       200  // 回路図のR1=200Ω  抵抗は出力が振り切れない範囲で大きい方が計測精度が良い
#define CH1_R_HIGH_RESOLUTION     3000  // 回路図のR3=3kΩ   実際の抵抗は誤差があるのでテスターで実測して
#define CH2_R_LOW_RESOLUTION       200  // 回路図のR2=200Ω  この値を変更すると精度が向上する
#define CH2_R_HIGH_RESOLUTION     3000  // 回路図のR4=3kΩ
#define SAMPLES 120    // 1サイクル(1/50秒)に何回サンプリングするか
                       // 大きいほど精度が高く計測できるが大きすぎるとＡ／Ｄ変換の速度が間に合わない 150あたりが限界

//---------------- 配線を変更したら変更する項目 -------------------------------------------
#define PORT_CH1_RANGE_SELECT_HIGH_RESOLUTION    3  //  Port Digital 3
#define PORT_CH1_RANGE_SELECT_LOW_RESOLUTION     2  //  Port Digital 2
#define PORT_CH2_RANGE_SELECT_HIGH_RESOLUTION   15  //  Port Analog  1
#define PORT_CH2_RANGE_SELECT_LOW_RESOLUTION    14  //  Port Analog  0

#define PORT_ZERO_CROSS 4  // Port Digital  4
#define PORT_CH1        18  // Port Analog   4
#define PORT_CH2        16  // Port Analog   2
#define PORT_REF        17  // Port Analog   3  
#define PORT_BUZZER     19  // Port Analog   5
#define PORT_LED        13  // Port Digital 13

LiquidCrystal lcd(5,6,7,8,9,10);

//---------------- 回路定数など -----------------------------------------------------------
#define SELECT          HIGH
#define NOT_SELECT      LOW

#define HIGH_RESOLUTION 0
#define LOW_RESOLUTION  1

#define REF_VOLTS    2.495              // TL431 Typ 2.495V
#define PERIOD  (1000000/FREQ/SAMPLES)  // microsec
const float pi = 3.141592;

int out_volts[SAMPLES];                 // 電流－＞電圧変換回路の出力の計測値
int ac100_sin_volts[SAMPLES];           // AC100V サインカーブ
int ref_volts_10bit;                    // TL431リファレンス電圧のA/D RAW値

//------------------------------------------------------------------------------------------
// ＡＣ１００Ｖサイン波形の生成 
// １／４サイクルでもよいがメモリー不足ではないので１サイクル分生成している
//------------------------------------------------------------------------------------------
void generate_ac100v_sine_curve()
{
  for( int time_index = 0 ; time_index < SAMPLES ; time_index++ ) {
    ac100_sin_volts[time_index] = abs(sin( ( 2 * pi ) / SAMPLES * time_index )) * sqrt(2) * 100; 
  }
}  

//------------------------------------------------------------------------------------------
// １０ビットの計測値を電流値[mA]に換算する
//
// INPUUT:
//   volt     -> 計測対象機器の10ビット計測値
//   ref_volt -> TL431のA/D計測値
//   R_range  -> 計測レンジ
//------------------------------------------------------------------------------------------
float convert_10bit_to_milli_amp( int volt, int ref_volt, int R_range )
{
  return ((float)volt * REF_VOLTS * CT_TURN) / ((float)ref_volt * (float)R_range);
} 

//------------------------------------------------------------------------------------------
// １サイクルぶんをサンプリングする
//
// INPUT:
//   port  -> arduinoアナログポート番号
//   range -> HIGH_RESOLUSION / LOW_RESOLUTION
//------------------------------------------------------------------------------------------
void ad_get( int port, int range )
{
  long time_next_period;
  long period = PERIOD;
 
  if( range == HIGH_RESOLUTION ) {
    if( port == PORT_CH1 ) {
      digitalWrite( PORT_CH1_RANGE_SELECT_HIGH_RESOLUTION, SELECT     );
      digitalWrite( PORT_CH1_RANGE_SELECT_LOW_RESOLUTION,  NOT_SELECT );
    } else {  // PORT_CH2
      digitalWrite( PORT_CH2_RANGE_SELECT_HIGH_RESOLUTION, SELECT     );
      digitalWrite( PORT_CH2_RANGE_SELECT_LOW_RESOLUTION,  NOT_SELECT );
    }
  } else {  // LOW_RESOLUTION
    if( port == PORT_CH1 ) {
      digitalWrite( PORT_CH1_RANGE_SELECT_HIGH_RESOLUTION, NOT_SELECT );
      digitalWrite( PORT_CH1_RANGE_SELECT_LOW_RESOLUTION,  SELECT     );
    } else {  // PORT_CH2
      digitalWrite( PORT_CH2_RANGE_SELECT_HIGH_RESOLUTION, NOT_SELECT );
      digitalWrite( PORT_CH2_RANGE_SELECT_LOW_RESOLUTION,  SELECT     );
    }
  }  
  delay(1); // デジタル回路が動くと電圧が変動したりするので1ms程度待っている。たぶん効果はほとんど無い。気分の問題。

  if( micros() > 4294927296 ) { // for内部でオーバーフローすると大小関係が崩れるのであふれる直前40ms前になったら
    while(micros() < 20000) {}  // オーバーフローするまで待つ
  }
  while(digitalRead(PORT_ZERO_CROSS) == LOW ) {}  // AC100Vの電圧がゼロになるまで待つ
  time_next_period = micros();
  for( int time_index = 0 ; time_index < SAMPLES ; time_index++ ) {
    out_volts[time_index] = analogRead( port );
    time_next_period += period;
    while( micros() < time_next_period ) {
//      digitalWrite( PORT_LED, HIGH );  // LEDが点灯しないならA/D変換速度が間に合っていないのでサンプル数を小さくすること
    }
  }  
}

//------------------------------------------------------------------------------------------
// リファレンス電圧のＡ／Ｄ変換１０ビット値を得る
// １回Ａ／Ｄ変換すれば済むが、Ａ／Ｄの揺らぎを考慮し１０回取得して平均値を求めている
// INPUT:
//   port  -> arduinoアナログポート番号 = TL431出力
//------------------------------------------------------------------------------------------
float get_ref_volt( int port )
{
  int total = 0;
  
  for( int t = 0 ; t < 10 ; t++ ) {
    delay(1);
    total += analogRead( port );
  }  

  return (float)total / 10;
}

//------------------------------------------------------------------------------------------
// 計測値のオフセット値を求める
// 機器の電流波形は正負均等に触れると見なし、平均値をオフセットとする 値は 大体 512 になる
//------------------------------------------------------------------------------------------
float calc_offset()
{
  long total = 0;

  for(int time_index = 0 ; time_index < SAMPLES ; time_index++ ) {  
    total += (long)out_volts[time_index];
  }

  return (float)total / SAMPLES;
}

//------------------------------------------------------------------------------------------
// １ビットあたりの電流値[mA]を求める
//
// INPUT:
//    range -> HIGH_RESOLUSION / LOW_RESOLUTION
//------------------------------------------------------------------------------------------
float get_amp_div( int port, int range )
{
  if( range == HIGH_RESOLUTION ) {
    if( port = PORT_CH1 ) {
      return (REF_VOLTS * CT_TURN) / ((float)ref_volts_10bit * CH1_R_HIGH_RESOLUTION) * 1000;
    } else { // CH2
      return (REF_VOLTS * CT_TURN) / ((float)ref_volts_10bit * CH2_R_HIGH_RESOLUTION) * 1000;
    }
  } else { // LOW_RESOLUTION
    if( port = PORT_CH1 ) {
      return (REF_VOLTS * CT_TURN) / ((float)ref_volts_10bit * CH1_R_LOW_RESOLUTION) * 1000;
    } else { // CH2
      return (REF_VOLTS * CT_TURN) / ((float)ref_volts_10bit * CH2_R_LOW_RESOLUTION) * 1000;
    }
  }
}

//------------------------------------------------------------------------------------------
// ブザーを０．２秒鳴らす
//------------------------------------------------------------------------------------------  
void beep()
{
  for( int i = 0 ; i < 100 ; i++ ) {
    digitalWrite( PORT_BUZZER, HIGH );
    delay(1);
    digitalWrite( PORT_BUZZER, LOW  );
    delay(1);
  }    
}

//------------------------------------------------------------------------------------------ 
// 消費電力の計算 ゼロクロス点からサンプリングされた波形であることを前提として電力を計算する
//------------------------------------------------------------------------------------------  
float calc_watt_zc( int port, int range )
{
  float offset;
  float total;
  float temp_volt;
  float amp_div;
  
  offset = calc_offset();

  amp_div = get_amp_div( port, range );
  total = 0.0;
  for( int time_index ; time_index < SAMPLES ; time_index++ ) {  // １サイクルぶん電力を積算
    temp_volt = abs((float)out_volts[time_index] - offset);
    total += (temp_volt * amp_div * (float)ac100_sin_volts[ time_index ] / 1000);
  }  

  return total / SAMPLES;
}

//------------------------------------------------------------------------------------------
// arduino setup
//------------------------------------------------------------------------------------------  
void setup()
{
  pinMode(PORT_CH1_RANGE_SELECT_HIGH_RESOLUTION, OUTPUT);
  pinMode(PORT_CH1_RANGE_SELECT_LOW_RESOLUTION,  OUTPUT);
  pinMode(PORT_CH2_RANGE_SELECT_HIGH_RESOLUTION, OUTPUT);
  pinMode(PORT_CH2_RANGE_SELECT_LOW_RESOLUTION,  OUTPUT);

  pinMode(PORT_LED,    OUTPUT);
  pinMode(PORT_BUZZER, OUTPUT);
  
  digitalWrite( PORT_CH1_RANGE_SELECT_HIGH_RESOLUTION, NOT_SELECT );
  digitalWrite( PORT_CH1_RANGE_SELECT_LOW_RESOLUTION,  SELECT     );
  digitalWrite( PORT_CH2_RANGE_SELECT_HIGH_RESOLUTION, NOT_SELECT );
  digitalWrite( PORT_CH2_RANGE_SELECT_LOW_RESOLUTION,  SELECT     );
  
  lcd.begin(8, 2);
  lcd.println("WatMeter");
  lcd.setCursor(0,1); 
  lcd.println("Sample 2");
  
  generate_ac100v_sine_curve();
  ref_volts_10bit = get_ref_volt( PORT_REF );
  
  beep();
  delay(500);
}


//------------------------------------------------------------------------------------------
// arduino loop
//------------------------------------------------------------------------------------------  
void loop()
{
  float watt_ch1;
  float watt_ch2;

  ad_get( PORT_CH1, LOW_RESOLUTION );
  watt_ch1 = calc_watt_zc( PORT_CH1, LOW_RESOLUTION ); 
  ad_get( PORT_CH2, LOW_RESOLUTION );
  watt_ch2 = calc_watt_zc( PORT_CH2, LOW_RESOLUTION );
  
  lcd.clear();
  lcd.print( "C1 " );
  lcd.print( (int)watt_ch1 );
  lcd.print( "W" );
  lcd.setCursor(0,1);  
  lcd.print( "C2 " );
  lcd.print( (int)watt_ch2 );
  lcd.print( "W" );  
  delay(300);
}

//------------------------------------------------------------------ E N D ------------------

